import json
import tarfile

import numpy as np
import pandas as pd

GROWTH_COLUMNS = [
    "TICK",
    "SEED",
    "U",
    "V",
    "W",
    "Z",
    "POSITION",
    "POPULATION",
    "STATE",
    "VOLUME",
    "CYCLE",
]

CELL_STATES = [
    "NEUTRAL",
    "APOPTOTIC",
    "QUIESCENT",
    "MIGRATORY",
    "PROLIFERATIVE",
    "SENESCENT",
    "NECROTIC",
]


def parse_growth_file(tar: tarfile.TarFile) -> pd.DataFrame:
    """
    Parses a tumor growth simulation tar file.

    Parameters
    ----------
    tar :
        Tar file of simulations for different seeds.

    Returns
    -------
    :
        Parsed simulation data for all seeds and timepoints.
    """

    all_timepoints = []

    for member in tar.getmembers():
        extracted_member = tar.extractfile(member)
        assert extracted_member is not None
        extracted_json = json.loads(extracted_member.read().decode("utf-8"))

        seed = extracted_json["seed"]
        all_timepoints.extend(
            [
                data
                for timepoint in extracted_json["timepoints"]
                for data in parse_growth_timepoint(timepoint, seed)
            ]
        )

    timepoints_df = pd.DataFrame(all_timepoints, columns=GROWTH_COLUMNS)

    return timepoints_df


def parse_growth_timepoint(timepoint: dict, seed: int) -> list:
    """
    Parses a simulation timepoint into a list of features per cell.

    The original data contains cell features in the form:

    .. code-block:: text

        {
            "time": time,
            "cells": [
                [
                    [u, v, w, z],
                    [
                        [
                            type,
                            population,
                            state,
                            position,
                            volume,
                            [cell, cycle, lengths, ...]
                        ],
                        ...
                    ]
                ],
                ...
            ]
        }

    Parsed data is formatted into:

    .. code-block:: text

        [
            [time, seed, u, v, w, z, position, population, state, volume, cycle],
            [time, seed, u, v, w, z, position, population, state, volume, cycle],
            ...
        ]

    Cell cycle length is `None` if the cell has not yet divided. Otherwise, cell
    cycle is the average of all cell cycle lengths.

    Parameters
    ----------
    timepoint :
        Data for a timepoint.

    Returns
    -------
    :
        Parsed data of the timepoint.
    """

    parsed_data = []
    time = timepoint["time"]

    for (location, cells) in timepoint["cells"]:
        u, v, w, z = location

        for cell in cells:
            _, population, state, position, volume, cycles = cell

            if len(cycles) == 0:
                cycle = None
            else:
                cycle = np.mean(cycles)

            data_list = [
                time,
                seed,
                u,
                v,
                w,
                z,
                position,
                population,
                CELL_STATES[state],
                volume,
                cycle,
            ]

            parsed_data.append(data_list)

    return parsed_data



import json
import tarfile
from typing import Optional, Union

import numpy as np


def extract_tick_json(
    tar: tarfile.TarFile,
    key: str,
    tick: Union[int, float],
    extension: Optional[str] = None,
    field: Optional[str] = None,
) -> list:
    formatted_tick = f"_{tick:06d}" if isinstance(tick, (int, np.integer)) else ""

    if extension is None:
        member = tar.extractfile(f"{key}{formatted_tick}.json")
    else:
        member = tar.extractfile(f"{key}{formatted_tick}.{extension}.json")

    assert member is not None
    tick_json = json.loads(member.read().decode("utf-8"))

    if isinstance(tick, float):
        tick_json = next(item for item in tick_json["timepoints"] if item["time"] == tick)[field]

    return tick_json



"""Tasks for converting ARCADE files to other formats."""

from prefect import task

from .convert_to_colorizer import convert_to_colorizer
from .convert_to_images import convert_to_images
from .convert_to_meshes import convert_to_meshes
from .convert_to_projection import convert_to_projection
from .convert_to_simularium import convert_to_simularium

convert_to_colorizer = task(convert_to_colorizer)
convert_to_images = task(convert_to_images)
convert_to_meshes = task(convert_to_meshes)
convert_to_projection = task(convert_to_projection)
convert_to_simularium = task(convert_to_simularium)



from typing import Optional


def get_location_voxels(location: dict, region: Optional[str] = None) -> list[tuple[int, int, int]]:
    voxels = [
        (x, y, z)
        for loc in location["location"]
        for x, y, z in loc["voxels"]
        if not region or loc["region"] == region
    ]
    return voxels



import itertools
import random
import tarfile
from math import cos, isnan, pi, sin, sqrt
from typing import Optional, Union

import numpy as np
import pandas as pd
from simulariumio import (
    DISPLAY_TYPE,
    AgentData,
    CameraData,
    DimensionData,
    DisplayData,
    MetaData,
    ModelMetaData,
    TrajectoryConverter,
    TrajectoryData,
    UnitData,
)

from arcade_collection.output.extract_tick_json import extract_tick_json
from arcade_collection.output.get_location_voxels import get_location_voxels

CELL_STATES: list[str] = [
    "UNDEFINED",
    "APOPTOTIC",
    "QUIESCENT",
    "MIGRATORY",
    "PROLIFERATIVE",
    "SENESCENT",
    "NECROTIC",
]

EDGE_TYPES: list[str] = [
    "ARTERIOLE",
    "ARTERY",
    "CAPILLARY",
    "VEIN",
    "VENULE",
    "UNDEFINED",
]


CAMERA_POSITIONS: dict[str, tuple[float, float, float]] = {
    "patch": (0.0, -0.5, 900),
    "potts": (10.0, 0.0, 200.0),
}

CAMERA_LOOK_AT: dict[str, tuple[float, float, float]] = {
    "patch": (0.0, -0.2, 0.0),
    "potts": (10.0, 0.0, 0.0),
}


def convert_to_simularium(
    series_key: str,
    simulation_type: str,
    data_tars: dict[str, tarfile.TarFile],
    frame_spec: tuple[int, int, int],
    box: tuple[int, int, int],
    ds: float,
    dz: float,
    dt: float,
    colors: dict[str, str],
    resolution: Optional[int] = None,
    url: Optional[str] = None,
) -> str:

    if simulation_type == "patch":
        frames = list(map(float, np.arange(*frame_spec)))
        radius, margin, height = box
        bounds = radius + margin
        length = (2 / sqrt(3)) * (3 * (radius + margin) - 1)
        width = 4 * (radius + margin) - 2
        data = format_patch_tar_data(
            series_key, data_tars["cells"], data_tars["graph"], frames, bounds
        )
    elif simulation_type == "potts":
        frames = list(map(int, np.arange(*frame_spec)))
        length, width, height = box
        data = format_potts_tar_data(
            series_key, data_tars["cells"], data_tars["locations"], frames, resolution
        )
    else:
        raise ValueError(f"invalid simulation type {simulation_type}")

    meta_data = get_meta_data(series_key, simulation_type, length, width, height, ds, dz)
    agent_data = get_agent_data(data)
    agent_data.display_data = get_display_data(series_key, data, colors, url)

    for index, (frame, group) in enumerate(data.groupby("frame")):
        n_agents = len(group)
        agent_data.times[index] = float(frame) * dt
        agent_data.n_agents[index] = n_agents
        agent_data.unique_ids[index][:n_agents] = range(0, n_agents)
        agent_data.types[index][:n_agents] = group["name"]
        agent_data.radii[index][:n_agents] = group["radius"]
        agent_data.positions[index][:n_agents] = group[["x", "y", "z"]]
        agent_data.n_subpoints[index][:n_agents] = group["points"].map(lambda points: len(points))
        agent_data.viz_types[index][:n_agents] = group["points"].map(
            lambda points: 1001 if points else 1000
        )
        agent_data.subpoints[index][:n_agents] = np.array(
            list(itertools.zip_longest(*group["points"], fillvalue=0))
        ).T

    agent_data.positions[:, :, 0] = (agent_data.positions[:, :, 0] - length / 2.0) * ds
    agent_data.positions[:, :, 1] = (width / 2.0 - agent_data.positions[:, :, 1]) * ds
    agent_data.positions[:, :, 2] = (agent_data.positions[:, :, 2] - height / 2.0) * dz

    agent_data.subpoints[:, :, 0::3] = (agent_data.subpoints[:, :, 0::3]) * ds
    agent_data.subpoints[:, :, 1::3] = (-agent_data.subpoints[:, :, 1::3]) * ds
    agent_data.subpoints[:, :, 2::3] = (agent_data.subpoints[:, :, 2::3]) * dz

    return TrajectoryConverter(
        TrajectoryData(
            meta_data=meta_data,
            agent_data=agent_data,
            time_units=UnitData("hr"),
            spatial_units=UnitData("um"),
        )
    ).to_JSON()


def get_meta_data(
    series_key: str,
    simulation_type: str,
    length: Union[int, float],
    width: Union[int, float],
    height: Union[int, float],
    ds: float,
    dz: float,
) -> MetaData:
    meta_data = MetaData(
        box_size=np.array([length * ds, width * ds, height * dz]),
        camera_defaults=CameraData(
            position=np.array(CAMERA_POSITIONS[simulation_type]),
            look_at_position=np.array(CAMERA_LOOK_AT[simulation_type]),
            fov_degrees=60.0,
        ),
        trajectory_title=f"ARCADE - {series_key}",
        model_meta_data=ModelMetaData(
            title="ARCADE",
            version=simulation_type,
            description=(f"Agent-based modeling framework ARCADE for {series_key}."),
        ),
    )

    return meta_data


def get_agent_data(data: pd.DataFrame) -> AgentData:
    total_frames = len(data["frame"].unique())
    max_agents = data.groupby("frame")["name"].count().max()
    max_subpoints = data["points"].map(lambda points: len(points)).max()
    return AgentData.from_dimensions(DimensionData(total_frames, max_agents, max_subpoints))


def get_display_data(
    series_key: str, data: pd.DataFrame, colors: dict[str, str], url: Optional[str] = None
) -> DisplayData:
    display_data = {}

    for name in data["name"].unique():
        group, cell_id, color_key, frame = name.split("#")

        random.seed(cell_id)
        jitter = (random.random() - 0.5) / 2

        if url is not None:
            display_data[name] = DisplayData(
                name=cell_id,
                display_type=DISPLAY_TYPE.OBJ,
                url=f"{url}/{series_key}_{int(frame):06d}_{int(cell_id):06d}_{group}.MESH.obj",
                color=shade_color(colors[color_key], jitter),
            )
        elif cell_id is None:
            display_data[name] = DisplayData(
                name=group,
                display_type=DISPLAY_TYPE.FIBER,
                color=colors[color_key],
            )
        else:
            display_data[name] = DisplayData(
                name=cell_id,
                display_type=DISPLAY_TYPE.SPHERE,
                color=shade_color(colors[color_key], jitter),
            )

    return display_data


def format_patch_tar_data(
    series_key: str,
    cells_tar: tarfile.TarFile,
    graph_tar: Optional[tarfile.TarFile],
    frames: list[Union[int, float]],
    bounds: int,
) -> pd.DataFrame:
    data: list[list[Union[int, str, float]]] = []

    theta = [pi * (60 * i) / 180.0 for i in range(6)]
    dx = [cos(t) / sqrt(3) for t in theta]
    dy = [sin(t) / sqrt(3) for t in theta]

    for frame in frames:
        cell_timepoint = extract_tick_json(cells_tar, series_key, frame, field="cells")

        for location, cells in cell_timepoint:
            u, v, w, z = location
            rotation = random.randint(0, 5)

            for cell in cells:
                _, population, state, position, volume, _ = cell
                cell_id = f"{u}{v}{w}{z}{position}"

                name = f"POPULATION{population}#{cell_id}#{CELL_STATES[state]}#"
                radius = (volume ** (1.0 / 3)) / 1.5

                x = (u + bounds - 1) * sqrt(3) + 1
                y = (v - w) + 2 * bounds - 1

                center = [
                    (x + dx[(position + rotation) % 6]),
                    (y + dy[(position + rotation) % 6]),
                    z,
                ]

                data = data + [[name, frame, radius] + center + [[]]]

        if graph_tar is not None:
            graph_timepoint = extract_tick_json(
                graph_tar, series_key, frame, "GRAPH", field="graph"
            )

            for (from_node, to_node, edge) in graph_timepoint:
                edge_type, radius, _, _, _, _, flow = edge

                name = f"VASCULATURE##{'UNDEFINED' if isnan(flow) else EDGE_TYPES[edge_type + 2]}#"

                subpoints = [
                    from_node[0] / sqrt(3),
                    from_node[1],
                    from_node[2],
                    to_node[0] / sqrt(3),
                    to_node[1],
                    to_node[2],
                ]

                data = data + [[name, frame, radius] + [0, 0, 0] + [subpoints]]

    return pd.DataFrame(data, columns=["name", "frame", "radius", "x", "y", "z", "points"])


def format_potts_tar_data(
    series_key: str,
    cells_tar: tarfile.TarFile,
    locations_tar: tarfile.TarFile,
    frames: list[Union[int, float]],
    resolution: Optional[int],
) -> pd.DataFrame:
    data: list[list[Union[int, str, float]]] = []

    for frame in frames:
        cells = extract_tick_json(cells_tar, series_key, frame, "CELLS")
        locations = extract_tick_json(locations_tar, series_key, frame, "LOCATIONS")

        for cell, location in zip(cells, locations):
            regions = [loc["region"] for loc in location["location"]]

            for region in regions:
                name = f"{region}#{cell['id']}#{cell['phase']}#"

                all_voxels = get_location_voxels(location, region if region != "DEFAULT" else None)

                if resolution is None:
                    radius = (len(all_voxels) ** (1.0 / 3)) / 1.5
                    center = list(np.array(all_voxels).mean(axis=0))
                    data = data + [[name, int(frame), radius] + center + [[]]]
                elif resolution == 0:
                    radius = 1
                    center = list(np.array(all_voxels).mean(axis=0))
                    data = data + [[f"{name}{frame}", int(frame), radius] + center + [[]]]
                else:
                    radius = resolution / 2
                    center_offset = (resolution - 1) / 2

                    resolution_voxels = get_resolution_voxels(all_voxels, resolution)
                    border_voxels = filter_border_voxels(set(resolution_voxels), resolution)
                    center_voxels = [
                        [x + center_offset, y + center_offset, z + center_offset]
                        for x, y, z in border_voxels
                    ]

                    data = data + [
                        [name, int(frame), radius] + voxel + [[]] for voxel in center_voxels
                    ]

    return pd.DataFrame(data, columns=["name", "frame", "radius", "x", "y", "z", "points"])


def get_resolution_voxels(
    voxels: list[tuple[int, int, int]], resolution: int
) -> list[tuple[int, int, int]]:
    df = pd.DataFrame(voxels, columns=["x", "y", "z"])

    min_x, min_y, min_z = df.min()
    max_x, max_y, max_z = df.max()

    samples = [
        (sx, sy, sz)
        for sx in np.arange(min_x, max_x, resolution)
        for sy in np.arange(min_y, max_y, resolution)
        for sz in np.arange(min_z, max_z, resolution)
    ]

    offsets = [
        (dx, dy, dz)
        for dx in range(resolution)
        for dy in range(resolution)
        for dz in range(resolution)
    ]

    resolution_voxels = []

    for sx, sy, sz in samples:
        sample_voxels = [(sx + dx, sy + dy, sz + dz) for dx, dy, dz in offsets]

        if len(set(sample_voxels) - set(voxels)) < len(offsets) / 2:
            resolution_voxels.append((sx, sy, sz))

    return resolution_voxels


def filter_border_voxels(
    voxels: set[tuple[int, int, int]], resolution: int
) -> list[tuple[int, int, int]]:
    offsets = [
        (resolution, 0, 0),
        (-resolution, 0, 0),
        (0, resolution, 0),
        (0, -resolution, 0),
        (0, 0, resolution),
        (0, 0, -resolution),
    ]
    filtered_voxels = []

    for x, y, z in voxels:
        neighbors = [(x + dx, y + dy, z + dz) for dx, dy, dz in offsets]
        if len(set(neighbors) - set(voxels)) != 0:
            filtered_voxels.append((x, y, z))

    return filtered_voxels


def shade_color(color: str, alpha: float) -> str:
    old_color = color.replace("#", "")
    old_red, old_green, old_blue = [int(old_color[i : i + 2], 16) for i in (0, 2, 4)]
    layer_color = 0 if alpha < 0 else 255

    new_red = round(old_red + (layer_color - old_red) * abs(alpha))
    new_green = round(old_green + (layer_color - old_green) * abs(alpha))
    new_blue = round(old_blue + (layer_color - old_blue) * abs(alpha))

    return f"#{new_red:02x}{new_green:02x}{new_blue:02x}"



import numpy as np
import pandas as pd

from arcade_collection.output.convert_model_units import convert_model_units


def convert_to_colorizer(
    all_data: pd.DataFrame,
    features: list[str],
    frame_spec: tuple[int, int, int],
    ds: float,
    dt: float,
    regions: list[str],
) -> dict:
    frames = list(np.arange(*frame_spec))
    manifest = get_manifest_data(features, frames)

    convert_model_units(all_data, ds, dt, regions)
    frame_data = all_data[all_data["TICK"].isin(frames)]

    outliers = get_outliers_from_data(frame_data)
    tracks = get_tracks_from_data(frame_data)
    times = get_times_from_data(frame_data)

    colorizer_json = {
        "manifest": manifest,
        "outliers": outliers,
        "tracks": tracks,
        "times": times,
    }

    for feature in features:
        colorizer_json[feature] = get_feature_from_data(frame_data, feature)

    return colorizer_json


def get_manifest_data(features: list[str], frames: list[int]) -> dict:
    manifest = {
        "frames": [f"frame_{i}.png" for i in range(len(frames))],
        "features": {
            feature: f"feature_{feature_index}.json"
            for feature_index, feature in enumerate(features)
        },
        "outliers": "outliers.json",
        "tracks": "tracks.json",
        "times": "times.json",
    }

    return manifest


def get_outliers_from_data(data: pd.DataFrame) -> dict:
    outliers = [False] * len(data)
    outliers_json = {"data": outliers, "min": False, "max": True}
    return outliers_json


def get_tracks_from_data(data: pd.DataFrame) -> dict:
    tracks = data["ID"]
    tracks_json = {"data": list(tracks)}
    return tracks_json


def get_times_from_data(data: pd.DataFrame) -> dict:
    times = data["time"]
    times_json = {"data": list(times)}
    return times_json


def get_feature_from_data(data: pd.DataFrame, feature: str) -> dict:
    feature_values = data[feature]
    feature_min = float(np.nanmin(feature_values))
    feature_max = float(np.nanmax(feature_values))

    feature_json = {"data": list(feature_values), "min": feature_min, "max": feature_max}
    return feature_json



"""Tasks for generating input files for ARCADE."""

from prefect import task

from .convert_to_cells_file import convert_to_cells_file
from .convert_to_locations_file import convert_to_locations_file
from .generate_setup_file import generate_setup_file
from .group_template_conditions import group_template_conditions
from .merge_region_samples import merge_region_samples

convert_to_cells_file = task(convert_to_cells_file)
convert_to_locations_file = task(convert_to_locations_file)
generate_setup_file = task(generate_setup_file)
group_template_conditions = task(group_template_conditions)
merge_region_samples = task(merge_region_samples)



import tarfile
from typing import Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle
from skimage import measure

from arcade_collection.output.extract_tick_json import extract_tick_json
from arcade_collection.output.get_location_voxels import get_location_voxels


def convert_to_projection(
    series_key: str,
    data_tar: tarfile.TarFile,
    frame: int,
    regions: list[str],
    box: tuple[int, int, int],
    ds: float,
    dt: float,
    scale: int,
    colors: dict[str, str],
) -> mpl.figure.Figure:
    fig = plt.figure(figsize=(10, 10), constrained_layout=True)
    length, width, height = box

    ax = fig.add_subplot()

    ax.invert_yaxis()
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])
    ax.set_xlim([0, length - 1])
    ax.set_ylim([width - 1, 0])
    ax.set_box_aspect(1)

    ax_horz = ax.inset_axes([0, 1.005, 1, height / width], sharex=ax)
    ax_horz.set_ylim([0, height - 1])
    ax_horz.get_yaxis().set_ticks([])

    ax_vert = ax.inset_axes([1.005, 0, height / length, 1], sharey=ax)
    ax_vert.set_xlim([0, height - 1])
    ax_vert.get_xaxis().set_ticks([])

    ax.set_facecolor("#000")
    ax_horz.set_facecolor("#000")
    ax_vert.set_facecolor("#000")

    locations = extract_tick_json(data_tar, series_key, frame, "LOCATIONS")

    for region in regions:
        color = colors[region]

        for location in locations:
            for contour in get_array_contours(location, length, width, height, region):
                ax.plot(contour[:, 0], contour[:, 1], linewidth=0.5, color=color, alpha=0.5)

            for contour in get_array_contours(location, length, width, height, region, (0, 2, 1)):
                ax_horz.plot(contour[:, 0], contour[:, 1], linewidth=0.5, color=color, alpha=0.5)

            for contour in get_array_contours(location, length, width, height, region, (2, 1, 0)):
                ax_vert.plot(contour[:, 0], contour[:, 1], linewidth=0.5, color=color, alpha=0.5)

    add_frame_timestamp(ax, length, width, dt, frame, "#ffffff")
    add_frame_scalebar(ax, length, width, ds, scale, "#ffffff")

    return fig


def get_array_contours(
    location: dict,
    length: int,
    width: int,
    height: int,
    region: Optional[str] = None,
    rotate: Optional[tuple[int, int, int]] = None,
) -> list[np.ndarray]:
    array = np.zeros((length, width, height))
    voxels = get_location_voxels(location, region)

    if len(voxels) == 0:
        return []

    array[tuple(np.transpose(voxels))] = 1

    if rotate is not None:
        array = np.moveaxis(array, [0, 1, 2], rotate)
        length, width, height = array.shape

    contours: list[np.ndarray] = []

    for z in range(1, height):
        array_slice = array[:, :, z]

        if np.sum(array_slice) == 0:
            continue

        contours = contours + measure.find_contours(array_slice)

    return contours


def add_frame_timestamp(
    ax: mpl.axes.Axes, length: int, width: int, dt: float, frame: int, color: str
) -> None:
    hours, minutes = divmod(round(frame * dt, 2), 1)
    timestamp = f"{int(hours):02d}H:{round(minutes*60):02d}M"

    ax.text(
        0.03 * length,
        0.96 * width,
        timestamp,
        fontfamily="monospace",
        fontsize=20,
        color=color,
        fontweight="bold",
    )


def add_frame_scalebar(
    ax: mpl.axes.Axes, length: int, width: int, ds: float, scale: int, color: str
) -> None:
    scalebar = scale / ds

    ax.add_patch(
        Rectangle(
            (0.95 * length - scalebar, 0.94 * width),
            scalebar,
            0.01 * width,
            snap=True,
            color=color,
        )
    )

    ax.text(
        0.95 * length - scalebar / 2,
        0.975 * width,
        f"{scale} $\\mu$m",
        fontsize=10,
        color=color,
        horizontalalignment="center",
    )



import tarfile

import numpy as np
from skimage import measure

from arcade_collection.output.extract_tick_json import extract_tick_json
from arcade_collection.output.get_location_voxels import get_location_voxels


def convert_to_meshes(
    series_key: str,
    data_tar: tarfile.TarFile,
    frame_spec: tuple[int, int, int],
    regions: list[str],
    box: tuple[int, int, int],
    invert: bool = False,
) -> list[tuple[int, int, str, str]]:
    frames = list(np.arange(*frame_spec))
    meshes = []

    length, width, height = box

    for frame in frames:
        locations = extract_tick_json(data_tar, series_key, frame, "LOCATIONS")

        for location in locations:
            location_id = location["id"]

            for region in regions:
                voxels = [
                    (x, width - y - 1, z)
                    for x, y, z in get_location_voxels(
                        location, region if region != "DEFAULT" else None
                    )
                ]

                if len(voxels) == 0:
                    continue

                center = list(np.array(voxels).mean(axis=0))
                array = make_mesh_array(voxels, length, width, height)
                mesh = make_mesh_object(array, center, invert)

                meshes.append((frame, location_id, region, mesh))

    return meshes


def make_mesh_array(
    voxels: list[tuple[int, int, int]], length: int, width: int, height: int
) -> np.ndarray:
    # Create array.
    array = np.zeros((length, width, height), dtype=np.uint8)
    array[tuple(np.transpose(voxels))] = 7

    # Get set of zero neighbors for all voxels.
    offsets = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1)]
    neighbors = {
        (x + i, y + j, z + k)
        for x, y, z in voxels
        for i, j, k in offsets
        if array[x + i, y + j, z + k] == 0
    }

    # Remove invalid neighbors on borders.
    neighbors = {
        (x, y, z)
        for x, y, z in neighbors
        if x != 0 and y != 0 and z != 0 and x != length - 1 and y != width - 1 and z != height - 1
    }

    # Smooth array levels based on neighbor counts.
    for x, y, z in neighbors:
        array[x, y, z] = sum(array[x + i, y + j, z + k] == 7 for i, j, k in offsets) + 1

    return array


def make_mesh_object(array: np.ndarray, center: list[float], invert: bool = False) -> str:
    verts, faces, normals, _ = measure.marching_cubes(array, level=3, allow_degenerate=False)
    mesh = ""
    faces = faces + 1

    for item in verts:
        mesh += f"v {item[0] - center[0]} {item[1] - center[1]} {item[2] - center[2]}\n"

    for item in normals:
        mesh += f"vn {item[0]} {item[1]} {item[2]}\n"

    for item in faces:
        if invert:
            mesh += f"f {item[0]}//{item[0]} {item[1]}//{item[1]} {item[2]}//{item[2]}\n"
        else:
            mesh += f"f {item[2]}//{item[2]} {item[1]}//{item[1]} {item[0]}//{item[0]}\n"

    return mesh



"""Tasks for processing output files from ARCADE."""

from prefect import task

from .convert_model_units import convert_model_units
from .extract_tick_json import extract_tick_json
from .get_location_voxels import get_location_voxels
from .merge_parsed_results import merge_parsed_results
from .parse_cells_file import parse_cells_file
from .parse_growth_file import parse_growth_file
from .parse_locations_file import parse_locations_file

convert_model_units = task(convert_model_units)
extract_tick_json = task(extract_tick_json)
get_location_voxels = task(get_location_voxels)
merge_parsed_results = task(merge_parsed_results)
parse_cells_file = task(parse_cells_file)
parse_growth_file = task(parse_growth_file)
parse_locations_file = task(parse_locations_file)
